# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


EXTENSION_TO_MIMETYPE = {
    ".wasm": "application/wasm",
    ".js": "text/javascript",
    ".css": "text/css",
}
