// This file is part of Xpra.
// Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
// Xpra is released under the terms of the GNU GPL v2, or, at your option, any
// later version. See the file COPYING for details.

int is_x11_display(void *display);
