/*
 * This file is part of Xpra.
 * Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
 * Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 * later version. See the file COPYING for details.
 *
 * Make the progress bar more visible
 * (but we can't set the min-width this way...)
 */

progress, trough {
  min-height: 30px;
}
