# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

display = True
windows = True
cursors = True
gstreamer = True
x11 = True
webcam = True
audio = True
clipboard = True
notifications = True
dbus = True
mmap = True
ssl = True
ssh = True
logging = True
tray = True
network_state = True
network_listener = True
encoding = True
